﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace Hilbert
{
    [Serializable, StructLayout( LayoutKind.Sequential )]
    public struct Vector2f : IEquatable< Vector2f >
    {
        private float x;
        private float y;

        public static Vector2f Zero
        {
            get
            {
                return new Vector2f( 0, 0 );
            }
        }

        public static Vector2f Up
        {
            get
            {
                return new Vector2f( 0, 1 );
            }
        }

        public static Vector2f Right
        {
            get
            {
                return new Vector2f( 1, 0 );
            }
        }

        public Vector2f( float[] v )
        {
            x = v[ 0 ];
            y = v[ 1 ];
        }

        public Vector2f( float x, float y )
        {
            this.x = x;
            this.y = y;
        }

        public Vector2f( Vector2f v )
            : this( v.x, v.y )
        {
            
        }

        public Vector2f( Vector2i v )
            : this( v.X, v.Y )
        {
            
        }

        public float X
        {
            get
            {
                return x;
            }
            set
            {
                x = value;
            }
        }

        public float Y
        {
            get
            {
                return y;
            }
            set
            {
                y = value;
            }
        }

        /// <summary>
        /// Component-wise absolute value
        /// </summary>
        /// <returns></returns>
        public Vector2f Abs()
        {
            return new Vector2f( Math.Abs( x ), Math.Abs( y ) );
        }

        public float Norm()
        {
            return ( float )( Math.Sqrt( x * x + y * y ) );

        }

        public float NormSquared()
        {
            return ( x * x + y * y );
        }

        public void Normalize()
        {
            float length = Norm();

            x = x / length;
            y = y / length;
        }

        public Vector2f Normalized()
        {
            float length = Norm();

            return new Vector2f
            (
                x / length,
                y / length
            );
        }

        /// <summary>
        /// Returns a Vector orthogonal to this.
        /// Its norm depends on the vector, but is zero only for the zero vector.
        /// Note that the function this function is not continuous.
        /// </summary>
        /// <returns></returns>
        public Vector2f OrthogonalVector()
        {
            return new Vector2f( -y, x );
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
        
        public override bool Equals( object o )
        {
            if( o is Vector2f )
            {                
                Vector2f rhs = ( Vector2f )o;
                return( Equals( rhs ) );
            }
            return false;
        }

        public bool Equals( Vector2f other )
        {
            return ( x == other.x && y == other.y );
        }

        public override String ToString()
        {
            return string.Format( "< {0}, {1} >", x, y );
        }        

        // explicit conversion to Vector2i
        public static explicit operator Vector2i( Vector2f v )
        {
            return new Vector2i( ( int )( v.x ), ( int )( v.y ) );
        }

        public static Vector2f operator + ( Vector2f lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.x + rhs.x, lhs.y + rhs.y );
        }

        public static Vector2f operator + ( Vector2f lhs, Vector2i rhs )
        {
            return new Vector2f( lhs.x + rhs.X, lhs.y + rhs.Y );
        }

        public static Vector2f operator + ( Vector2i lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.X + rhs.x, lhs.Y + rhs.y );
        }

        public static Vector2f operator - ( Vector2f v )
        {
            return new Vector2f( -v.x, -v.y );
        }

        public static Vector2f operator - ( Vector2f lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.x - rhs.x, lhs.y - rhs.y );
        }

        public static Vector2f operator - ( Vector2f lhs, Vector2i rhs )
        {
            return new Vector2f( lhs.x - rhs.X, lhs.y - rhs.Y );
        }

        public static Vector2f operator - ( Vector2i lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.X - rhs.x, lhs.Y - rhs.y );
        }

        public static Vector2f operator * ( float a, Vector2f rhs )
        {
            return new Vector2f( a * rhs.x, a * rhs.y );
        }

        public static Vector2f operator * ( Vector2f lhs, float a )
        {
            return( a * lhs );
        }

        /// <summary>
        /// Per-element multiplication
        /// </summary>
        /// <param name="lhs"></param>
        /// <param name="rhs"></param>
        /// <returns></returns>
        public static Vector2f operator * ( Vector2f lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.x * rhs.x, lhs.y * rhs.y );
        }

        public static Vector2f operator / ( Vector2f v, float f )
        {
            float a = 1 / f;
            return new Vector2f( a * v.x, a * v.y );
        }

        public static Vector2f operator / ( Vector2f lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.x / rhs.x, lhs.y / rhs.y );
        }

        public static Vector2f operator / ( Vector2f lhs, Vector2i rhs )
        {
            return new Vector2f( lhs.x / rhs.X, lhs.y / rhs.Y );
        }

        public static Vector2f operator / ( Vector2i lhs, Vector2f rhs )
        {
            return new Vector2f( lhs.X / rhs.x, lhs.Y / rhs.y );
        }

        public static float Dot( Vector2f lhs, Vector2f rhs )
        {
            return( lhs.x * rhs.x + lhs.y * rhs.y );
        }
       
        public static Vector2f Lerp( Vector2f x, Vector2f y, float t )
        {
            return( x + t * ( y - x ) );
        }
    }
}
